/*
 * $Header: /cvsroot/xwinx/xwinx/xwinx.c,v 1.31 2001/11/16 07:22:12 snaund Exp $
 * xwinx.c - X11 client to allow remote access to Win32 desktop
 * Copyright (C) 2001 Kim Saunders, Sanjay Rao
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <wingdi.h>
#include <getopt.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <pthread.h>
#include "xwx.h"
#include "xwxmemory.h"
#include "xwxmouse.h"
#include "xwxprint.h"
#include "xwxkeyboard.h"
#include "xwxhostname.h"
#include "xwxhttp.h"
#include "xwxlines.h"
#include "xwx8bit.h"
#include "xwxinput.h"

/* use &lock for variables, or whatever else, in particular for
 * reading and writing "paused" and "toexit"
 */
static pthread_mutex_t lock;
static int paused;
static int toexit;

int ispaused_xwinx(void) 
{
	int i;

	pthread_mutex_lock(&lock);
	i = paused;
	pthread_mutex_unlock(&lock);
	return(i);
}

int pause_xwinx(void) 
{
	pthread_mutex_lock(&lock);
	paused = 1;
	pthread_mutex_unlock(&lock);
	return(0);
}

int unpause_xwinx(void) 
{
	pthread_mutex_lock(&lock);
	paused = 0;
	pthread_mutex_unlock(&lock);
	return(0);
}

void colorcorrect(char *oneline, int xpix) 
{
	unsigned short a;
	unsigned short b;
	unsigned short *shorts;
	int i;

	shorts=(unsigned short *)oneline;

	/* windows has 0RRRRRGGGGGBBBBB
		while X11 seems to have RRRRRGGGGGGBBBBB
	*/

	for (i=0;i<xpix;i++) {
		a=shorts[i];
		b=a&63;
		a=a<<1;
		a=(a&~127) | b;
		shorts[i]=a;
	}
}

static int get_params(struct params *params, int argc, char **argv) 
{
	int cop;	/* current option character */
	int idx;	/* option index */
	int toret=0; 	/* value to return, !0 => exit cleanly */

	static struct option longopts[] = {
		{"display",		1, 0, 'd'},
		{"mit-magic-cookie-1", 	1, 0, 'c'},
		{"bpp",			1, 0, 'b'},
		{"help",		0, 0, 'h'},
		{"version",		0, 0, 'v'},
		{"port",		1, 0, 'p'},
		{"pause",		1, 0, '0'},
		{"unpause",		1, 0, '1'},
		{"auth",		1, 0, 'a'},
		{"nodisplay",		0, 0, 'n'},
		{0,			0, 0, 0}
	};

	/* set the invocation name to the name used by the shell */
	params->invname = argv[0];
	params->cookie_len = 0;

	/* loop through all the command line options */
	while (1)
	{
		int i;	/* counter for reading cookie */

		/* set cop to the current option character */
		cop = getopt_long(
			argc, argv, "d:c:b:hvp:0:1:a:n", longopts, &idx
		);

		/* break out of the loop if there aren't any more options */
		if (cop == -1)
			break;

		switch (cop)
		{
			case 'd':
				params->display = strdup(optarg);
				break;
			case 'c':
				/* MIT-MAGIC-COOKIE-1 cookies must be 128 bits
				 * (32 chars) long
				 */
				if (strlen(optarg) != 32)
					errX11Cookie(params);
				
				params->cookie_len = 16;
				params->cookie_dat = (char *)xwxmalloc(
					params->cookie_len
				);

				/* this steps through the argument (a string)
				 * and sets each byte in cookie_dat to the
				 * *numerical value* of the corresponding
				 * 2 characters in the argment string
				 */
				for (i = 0; i < 16; i++)
				{
					int m;	/* number of matches */
					unsigned int byte;
					
					/* read 2 chars into byte */
					m = sscanf(
						&optarg[i * 2],
						"%2x",
						&byte
					);
					params->cookie_dat[i] = byte & 0xFF;

					/* makes sure the characters were valid
					 * hex (0-9A-F).
					 */
					if (m != 1)
						errX11Cookie(params);
				}

				break;
			case 'b':
				params->depth = atoi(optarg);
				break;
			case 'h':
				prnHelp(params);
				toret = 1; /* exit after printing */
				break;
			case 'v':
				prnVersion(params);
				toret = 1; /* exit after printing */
				break;
			case 'p':
				params->httpport = atoi(optarg);
				break;
			case '0':
				pause_httpd( atoi(optarg) );
				toret = 1; /* exit after action */
				break;
			case '1':
				unpause_httpd( atoi(optarg) );
				toret = 1; /* exit after action */
				break;
			case 'a':
				params->httpauth = strdup(optarg);
				break;
			case 'n':
				params->nodisplay = 1;
				break;
			case ':':
				/* invalid param => don't start */
				toret = 1;
				break;
		}
	}

	/* if we're going to exit, let's exit now */
	if (toret)
		return(toret);
	
	if (!params->display)
		params->display = getenv("DISPLAY");
	if (!params->display)
		params->display = "";

	prnHttpWarnings(params);

	return(0);
}

static int init_win32(struct params *params, struct win32info *win32info,
		struct imginfo *imginfo) 
{
	/* get the size of the windows desktop */
	win32info->dev = GetDC(NULL);
	imginfo->width = GetDeviceCaps(win32info->dev, HORZRES);
	imginfo->height = GetDeviceCaps(win32info->dev, VERTRES);

	win32info->mdev=CreateCompatibleDC(win32info->dev);
	if (!win32info->mdev)
		goto error;

	win32info->hbitmap = CreateCompatibleBitmap(
		win32info->dev,
		imginfo->width,
		imginfo->height
	);
	
	if (!win32info->hbitmap)
		goto error;

	win32info->obitmap = SelectObject(win32info->mdev, win32info->hbitmap);

	/* Credit to the VNC vncDesktop.cpp for these next couple lines */
	if ((GetDeviceCaps(win32info->dev, RASTERCAPS) & RC_DI_BITMAP) == 0)
	{
		errGetDIBits(params);
		return -1;
	}
	
	if (GetDeviceCaps(win32info->dev, PLANES) != 1)
	{
		errGetDIBits(params);
		return -1;
	}

	return 0;
	error:
		return -1;
}

static int init_x11win(
	struct params *params,
	struct x11info *x11info,
	struct imginfo *imginfo
) 
{
	XTextProperty titlepty; /* window title text property */
	char *title;
	char *hostname;

	x11info->attributes.background_pixel = BlackPixel(
		x11info->display,
		x11info->screen
	);
	
	x11info->attributes.border_pixel = WhitePixel(
		x11info->display,
		x11info->screen
	);
	
	/* create a window on the X server */
	x11info->window = XCreateWindow(
		x11info->display,
		DefaultRootWindow(x11info->display),
		0,
		0,
		imginfo->width,
		imginfo->height,
		0,
		imginfo->x11depth,
		InputOutput,
		x11info->visual,
		CWBackPixel | CWBorderPixel,
		&x11info->attributes
	);

	if (!x11info->window)
	{
		errX11Window(params);
		return(-1);
	}

	x11info->colormap = DefaultColormap(x11info->display, x11info->screen);
	
	if (imginfo->x11depth == 8)
	{
		if (mkcolors_8bit(x11info))
			return(-1);
	}

	/* get the hostname as a string */
	hostname = xwxhostname_gethostname();

	/* allocate enought space for the title string */
	title = (char *)xwxmalloc(
		strlen(XWX_NAME) + strlen(": ") + strlen(hostname) + 1
	);
	
	/* build the title string */
	sprintf(title, "%s: %s", XWX_NAME, hostname);

	/* set up the XTextProperty structure to contrain the title string */
	titlepty.encoding	= XA_STRING;
	titlepty.value		= title;
	titlepty.format		= 8;
	titlepty.nitems		= strlen(title);

	/* set the window name property */
	XSetWMName(x11info->display, x11info->window, &titlepty);

	/* get mouse click and keyboard button events */
	XSelectInput(
		x11info->display,
		x11info->window,
		ExposureMask |
		ButtonPressMask |
		ButtonReleaseMask |
		KeyPressMask |
		KeyReleaseMask |
		FocusChangeMask |
		PointerMotionMask
	);

	XMapWindow(x11info->display, x11info->window);
	
	/* get a structure containing the modifier <--> keysym mappings */

	x11info->modifiers = XGetModifierMapping(x11info->display);

	x11info->context = XCreateGC(
		x11info->display,
		x11info->window,
		0,
		&x11info->vls
	);

	XFlush(x11info->display);

	return(0);
}

static int init_x11(
	struct params *params,
	struct x11info *x11info,
	struct imginfo *imginfo
) 
{
	if (params->cookie_len)
		XSetAuthorization(
			"MIT-MAGIC-COOKIE-1",
			strlen("MIT-MAGIC-COOKIE-1"),
			params->cookie_dat,
			params->cookie_len
		);

	/* connect to X11 server */
	x11info->display = XOpenDisplay(params->display);
	if (!x11info->display)
	{
		errX11Display(params);
		return(-1);
	}

	x11info->screen = DefaultScreen(x11info->display);
	x11info->visual = DefaultVisual(x11info->display, x11info->screen);

	/* get default depth of X server */
	if (!params->depth)
		params->depth = DefaultDepth(x11info->display, x11info->screen); 
	switch (params->depth) {
		case 8:
			imginfo->x11depth = 8;
			imginfo->w32depth = 16;
			imginfo->Bpp = 1;
			imginfo->badcolor = 2;
			if (!(imginfo->width&3))
				imginfo->x11quantum = 32;
			else if (!(imginfo->width&1))
				imginfo->x11quantum = 16;
			else
				imginfo->x11quantum = 8;
			break;
		case 16:
			imginfo->x11depth = 16;
			imginfo->w32depth = 16;
			imginfo->Bpp = 2;
			imginfo->badcolor = 1;
			if (!(imginfo->width&1))
				imginfo->x11quantum = 32;
			else
				imginfo->x11quantum = 16;
			break;
		case 24:
			imginfo->x11depth = 24;
			imginfo->w32depth = 32;
			imginfo->Bpp = 4;
			imginfo->x11quantum = 32;
			break;
		default:
			errX11Depth(params);
			return(-1);
			break;
	}

	return 0;
}

/* set up XImage and win32 bitmaps */
static int init_img(
	struct params *params,
	struct x11info *x11info,
	struct imginfo *imginfo,
	struct win32info *win32info
)
{
	/* bytes tper line = width * bytes per pixel */
	imginfo->Bpl = imginfo->width * imginfo->Bpp;

	/* allocate memory for the (bytes = height * bytes per line) */ 
	imginfo->ximagebuffer = xwxmalloc(imginfo->height * imginfo->Bpl);

	/* make an image to put the screendump into */
	x11info->image = XCreateImage(
		x11info->display,
		x11info->visual,
		imginfo->x11depth,
		ZPixmap,
		0,
		imginfo->ximagebuffer,
		imginfo->width,
		imginfo->height,
		imginfo->x11quantum,
		0
	);

	/* print an error if the XImage wasn't created successfuly */
	if (!x11info->image)
	{
		errX11XImage(params);
		return -1;
	}

	/* this sets the rest of the XImage structure values from the members
	 * we have intialised
	 */	
	XInitImage(x11info->image);

	/* Bitmap header describing the bitmap we want to get from GetDIBits */
	win32info->bmi.bmiHeader.biSize		= sizeof(BITMAPINFOHEADER);
	win32info->bmi.bmiHeader.biWidth		= imginfo->width;
	win32info->bmi.bmiHeader.biHeight		= imginfo->height;
	win32info->bmi.bmiHeader.biPlanes		= 1;
	win32info->bmi.bmiHeader.biBitCount		= imginfo->w32depth;
	win32info->bmi.bmiHeader.biCompression		= BI_RGB;
	win32info->bmi.bmiHeader.biSizeImage		= 0;
	win32info->bmi.bmiHeader.biXPelsPerMeter	= 0;
	win32info->bmi.bmiHeader.biYPelsPerMeter	= 0;
	win32info->bmi.bmiHeader.biClrUsed		= 0;
	win32info->bmi.bmiHeader.biClrImportant		= 0;

	return 0;
}

static void deinit_win32(struct win32info *win32info) {
	/* TODO fill this in */
}

/* closes the connection to the X display */
static void deinit_x11(struct x11info *x11info) 
{
	/* TODO complete this */
	if (x11info->display) {
		XCloseDisplay(x11info->display);
		x11info->display = NULL;
	}
}

/* frees memory allocated for XImage */
static void deinit_img(struct imginfo *imginfo) 
{
	freeup(imginfo->ximagebuffer);
}

static int init_globals(void) 
{
	pthread_mutex_init(&lock, NULL);
	return 0;
}


static void deinit_globals(void)
{
	pthread_mutex_unlock(&lock);
	pthread_mutex_destroy(&lock);
}

static int nodisplay_main(struct params *params,
		struct imginfo *imginfo,
		struct win32info *win32info,
		struct x11info *x11info,
		struct xmission *xmission,
		struct xinput *xinput)
{

	if (!params->httpport)
		return(EXIT_SUCCESS);
	
	if (init_globals())
		goto error;

	/* start the http server thread */
	if (init_httpd(params, params->httpport))
		goto error;
	
	/* print XWinX banner with version and copyright info */
	prnBanner(params);

	while (1)
	{
		Sleep(3000);
		if (toexit)
			break;
	}
	
	deinit_httpd();
	deinit_globals();

	return(EXIT_SUCCESS);
error:
	fprintf(stderr, "An error was encountered, exiting...\n");
	return(EXIT_FAILURE);
}

static int display_main(
	struct params *params,
	struct imginfo *imginfo,
	struct win32info *win32info,
	struct x11info *x11info,
	struct xmission *xmission,
	struct xinput *xinput
)
{
	if (init_globals())
		goto error;

	/* set up connection to X display */
	if (init_x11(params, x11info, imginfo))
		goto error;
	
	if (init_win32(params, win32info, imginfo)) goto error;

	/* start the http server thread */
	if (init_httpd(params, params->httpport))
		goto error;
	
	if (init_x11win(params, x11info, imginfo))
		goto error;

	/* set up XImage and win32 bitmap */
	if (init_img(params, x11info, imginfo, win32info))
		goto error;

	/* print XWinX banner with version and copyright info */
	prnBanner(params);

	/* start a thread to handle mouse/keyboard events */
	if (init_input(params, x11info, xinput))
		goto error;

	/* from now on, don't use any X11 calls that aren't wrapped
	with xinput.inputlock */
	
	/* main loop, to detect graphical changes and transmit them */
	if (init_xmission(params, x11info, imginfo, win32info, xmission))
		goto error;

	/* transmit screen to X11 window */
	loop_xmission(params, x11info, imginfo, win32info, xmission, xinput);

	/* close connections, free memory, etc */
	deinit_xmission(xmission);
	deinit_input(xinput);
	deinit_img(imginfo);
	deinit_x11(x11info);
	deinit_httpd();
	deinit_win32(win32info);
	deinit_globals();
	return(EXIT_SUCCESS);

error:
	fprintf(stderr, "An error was encountered, exiting...\n");
	return(EXIT_FAILURE);
}


int main(int argc, char **argv) 
{
	struct params params;
	struct imginfo imginfo;
	struct win32info win32info;
	struct x11info x11info;
	struct xmission xmission;
	struct xinput xinput;

	/* ititialise all structures */
	memset(&imginfo, 0, sizeof(struct imginfo));
	memset(&params, 0, sizeof(struct params));
	memset(&win32info, 0, sizeof(struct win32info));
	memset(&x11info, 0, sizeof(struct x11info));
	memset(&xmission,0, sizeof(struct xmission));
	memset(&xinput, 0, sizeof(struct xinput));

	/* get command line options and parameters */
	if (get_params(&params, argc, argv))
		return 0;

	if (params.nodisplay)
		return nodisplay_main(
			&params,
			&imginfo,
			&win32info,
			&x11info,
			&xmission,
			&xinput
		);

	return display_main(
		&params,
		&imginfo,
		&win32info,
		&x11info,
		&xmission,
		&xinput
	);
}

