/*
 * $Header: /cvsroot/xwinx/xwinx/xwxlines.c,v 1.18 2001/11/08 21:59:47 guilty Exp $
 * xwxlines.c - transmission of line data for xwinx.c
 * Copyright (C) 2001 Kim Saunders, Sanjay Rao
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <wingdi.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <pthread.h>
#include "xwx.h"
#include "xwxmemory.h"
#include "xwxlines.h"
#include "xwxmouse.h"
#include "xwxkeyboard.h"
#include "xwxprint.h"


void deinit_xmission(struct xmission *xmission) 
{
	freeup(xmission->ldiff);
	freeup(xmission->cdiff);
	freeup(xmission->lefts);
	freeup(xmission->rights);
	freeup(xmission->oneline);
}

/* initializes the screen change detection and transmission code
 * currently this portion also polls for X events
 * this sets up vars for loop_xmission()
 */
int init_xmission(
		struct params *params,
		struct x11info *x11info,
		struct imginfo *imginfo,
		struct win32info *win32info,
		struct xmission *xmission
)
{
	/* 
	 * Note that imginfo->Bpp is for the X11 depth,
	 * which differs from the Win32 Bpp (imginfo->w32depth) in
	 * the 8bit case, where Bpp=1 but w32depth=2.
	 */
	switch (imginfo->w32depth) {
		case 16:
			/* +2 guarantees that the memory has 32bit padding
			 * (in case imginfo->width is odd)
			 */
			xmission->oneline = xwxmalloc(imginfo->width * 2 + 2);
			break;
		case 32:
			/* 32bit padding is quite unnecessary */
			xmission->oneline = xwxmalloc(imginfo->width * 4);
			break;
		default:
			fprintf(stderr, "w32depth not supported\n");
			xmission->oneline = NULL;
			break;
	}
	if (!xmission->oneline)
		goto error;

	/* ldiff stores an int of status for a line. */
	xmission->ldiff = xwxcalloc(1, sizeof(int) * (imginfo->height + 1)); 
		/* +1 to simplify alg in line decay */
	if (!xmission->ldiff)
		goto error;
	/* cdiff[] saves which lines have changed in an update cycle */
	xmission->cdiff=xwxcalloc(1, imginfo->height); 
	if (!xmission->cdiff)
		goto error;
	/* lefts[] and rights[] save the left-most and right-most element
	 * that is changed for a line. */
	xmission->lefts = xwxmalloc(sizeof(int) * imginfo->height);
	xmission->rights = xwxmalloc(sizeof(int) * imginfo->height);

	return(0);
	
error:
	return(-1);
}

/* this is for 8 bit X displays.
 * it looks up 15bit win32 colors and spits out 
 * 8bit X indexed color values */
static void colorlookup(char *oneline, int xpix, char *table)
{
	int i;
	unsigned short *temp;

	temp = (unsigned short*)oneline;

	for (i = 0; i < xpix; i++)
	{
		oneline[i] = table[*temp];
		temp++;
	}
}

/* loop_xmission is the guts of the screen updating
 * it searches for changed lines and when it finds them, bundles
 * them together somewhat and sends them to the x server.
 * it currently also polls for X events */
void loop_xmission(
		struct params *params_in,
		struct x11info *x11info_in,
		struct imginfo *imginfo_in,
		struct win32info *win32info_in,
		struct xmission *xmission_in,
		struct xinput *xinput
) 
{
	int i;
	int ypixd2;
	struct params params;
	struct x11info x11info;
	struct imginfo imginfo;
	struct win32info win32info;
	struct xmission xmission;
	int *lefts,*rights; /* saves info on the first and last change
				in the lines (lefts[i]==first change in row i)
				NOTE this is only valid for cdiff[i]==0
			    */
	int *ldiff; /* this is the memory of the status of the line
		       ldiff[i]==
		       		0 => update now
				<= 30 => update quickly
				<= 90 => update slowly
		       */
	char *oneline;
	char *cdiff; /* cdiff[i] == 0 => line i has changed
			cdiff[i] == 1 => line i was not checked
			cdiff[i] == 2 => line i was checked and didn't change
			*/
	char *xib; /* current snapshot of the screen */
	int xpix,ypix,Bpl,Bpp; /* x11 values */
	unsigned char *colorindex;
	int def_tickwait, slw_tickwait, tickwait;
	DWORD tick1,tick2;
	int back=0;

	/* I don't really know the theory, so if you know the proper
	 * way to interlace, please comment on this order 
	 *
	 * The idea is to check only 1/32 of the rows per cycle, to
	 * lower cpu use. Skipping every 32 rows per cycle makes sense,
	 * but to then move the offset by only 1 or 2 does not, as it would
	 * then ignore the 30 rows inbetween. Thus, we try to interlace
	 * the polling. So, a character drawing would change ~10 rows, and
	 * would be picked up most likely by the 3rd cycle, which would then
	 * search up and down for the end of the changes.
	 *
	 * interlace[] defines the order of the interlacing for a 32-row
	 * block
	 */
	char interlace[]={ 
		0, 16, 
		8, 24, 
		4, 20, 12, 28, 
		2, 18, 10, 26, 6, 22, 14, 30,
		1, 17,
		9, 25,
		5, 21, 13, 29,
		3, 19, 11, 27, 7, 23, 15, 31 };
	int cur_ilace;
	int idx_ilace=0;

	memcpy(&params, params_in, sizeof(struct params));
	memcpy(&x11info, x11info_in, sizeof(struct x11info));
	memcpy(&win32info, win32info_in, sizeof(struct win32info));
	memcpy(&imginfo, imginfo_in, sizeof(struct imginfo));
	memcpy(&xmission, xmission_in, sizeof(struct xmission));

	lefts	= xmission.lefts;
	rights	= xmission.rights;
	ldiff	= xmission.ldiff;
	oneline	= xmission.oneline;
	cdiff	= xmission.cdiff;
	xib	= imginfo.ximagebuffer;
	xpix	= imginfo.width;
	ypix	= imginfo.height;
	Bpl	= imginfo.Bpl;
	Bpp	= imginfo.Bpp;

	colorindex = x11info.colorindex;
	
	ypixd2 = ypix / 2;

	def_tickwait=30;
	slw_tickwait=1500;

	tickwait=def_tickwait;

	while (1)
	{
		int changes;
		int left, right;

		/* current ms count -- loops around */
		tick1=GetTickCount();

			/* check for exposure in xinput.c 
			 * check for focus at the same time
			 * */
		{
			int exp_x,exp_y,exp_w,exp_h,exp_b;

			pthread_mutex_lock(&xinput->inputlock);

			/* focus part here */
			if (xinput->unfocus) tickwait=slw_tickwait;
			else tickwait=def_tickwait;

			exp_b=xinput->exp_b;
			if (exp_b) {
				exp_x=xinput->exp_x;
				exp_y=xinput->exp_y;
				exp_w=xinput->exp_w;
				exp_h=xinput->exp_h;
				xinput->exp_b=0;

				/* already locked for this */
				XPutImage(
					x11info.display,
					x11info.window,
					x11info.context,
					x11info.image,
					exp_x,
					exp_y,
					exp_x,
					exp_y,
					exp_w,
					exp_h
				);

				/* undeadlock xinput in case it's waiting for us
				 * to process this event */
				SetEvent(xinput->deadlock);
			} 
			pthread_mutex_unlock(&xinput->inputlock);

		}
		
		
		/* line change detection here */
		changes = 0;
		if (!ispaused_xwinx())
		{

			/* move the interlace index since the last
			 * cycle */
			if (idx_ilace==31) idx_ilace=0;
			else idx_ilace++;

			cur_ilace=interlace[idx_ilace];
			
			/* set the "cdiff" for all lines to 1 => unchecked */
			memset(cdiff, 1, ypix);

			/* i is the current screen row we're considering */
			for (i = ypix-1; i >= 0; i--)
			{

				char *temp;

				/* i can go backwards, so if this line
				 * has already been checked then skip it */
				if (cdiff[i]!=1) continue;

				if (back) {
					/* this is from a backtrack --
					 * now we skip forward a line
					 * afterward 
					 *
					 * note that cdiff[i+1]==0
					 * */
					back=0;
				} else {
					/* see ldiff[] desc above */
					switch (ldiff[i]) {
						case 1:
							/* it _was_ a priority,
							 * but now it may not be */
						case 31:
							/* not a priority, still
							 * isn't */
							ldiff[i]=90;
							break;
						case 0:
							/* make it a priority */
							ldiff[i]=30;
							break;
						default:
							ldiff[i] -= 1;
							break;
					}


					if (ldiff[i]>30) {
						if ((i&31)!=cur_ilace) continue;
					}
				}
				
				/* this gets the line from windows */

#if 0
				if (!BitBlt(
					win32info.mdev,
					0,ypix-i-1,xpix,1,
					win32info.dev,
					0,ypix-i-1,SRCCOPY)) 
#endif
				if (!BitBlt(
					win32info.mdev,
					0,i,xpix,1,
					win32info.dev,
					0,i,SRCCOPY)) 
				{
					DWORD err;
					err=GetLastError();
					fprintf(stderr,"BitBlt "\
						"GetLastError()=%lu\n",
						err);
					errGetDIBits(params_in);
					goto error;
				}

				if (!GetDIBits(
					win32info.mdev,
					win32info.hbitmap,
					ypix - i - 1,
					1,
					oneline,
					&win32info.bmi,
					DIB_RGB_COLORS
				)) {
					DWORD err;
					err=GetLastError();
					fprintf(stderr,"GetDIBits "\
						"GetLastError()=%lu\n",
						err);
					errGetDIBits(params_in);
					goto error;
				} 
				
				/* win32 -> X11 has some color issues
				 * rewrite the colors in oneline[] 
				 * if necessary */
				if (imginfo.badcolor == 1)
					colorcorrect(oneline, xpix);
				else if (imginfo.badcolor == 2)
					colorlookup(oneline, xpix, colorindex);
				
				temp = xib + i * Bpl;

				if (!memcmp(oneline, temp, Bpl)) {
					/* line i hasn't changed since last update */
					cdiff[i]=2;
					continue;
				}

				/* find the left-most change in this line */
				for (
					left = 0;
					temp[left] == oneline[left];
					left++
				);

				/* find the right-most change in this line */
				for (
					right = Bpl - 1;
					temp[right] == oneline[right];
					right--
				);

				/* the changes above were byte values,
				 * not pixel values -- so convert it here */
				if (Bpp == 2)
				{
					left = left >> 1;
					right += 1;
					right = right >> 1;
				} else if (Bpp == 4)
				{
					/* Bpp == 4 */
					left = left >> 2;
					right += 3;
					right = right >> 2;
				}
				lefts[i] = left;
				rights[i] = right;
				memcpy(temp, oneline, Bpl);

				/* this line has changed */
				changes++;
				cdiff[i] = 0;


				if ((i)&&(ldiff[i - 1] > 30))
				{
					/* wake up the next line */
					ldiff[i - 1] = 0;
				}
				
				/* give this line priority */
				ldiff[i] = 0;

				/* wake up the previous line */
				/* ldiff[i+1] always exists */
				if (cdiff[i+1]==1) {
					/* mark that we're going back
					 */
					back=1;
					/* actually go back on the
					 * index variable !!! */
					i+=2;
				}
				/* WARNING do not use i below this point
				 * as it may be a different index */
			}
		}

		/* we've looked at the lines now -- let's send them */
		
		if (changes)
		{
			int start, last;

			for (i = 0; i < ypix; i++) 
				if (!cdiff[i]) break;

			start = i;
			last = i;

			if (i!=ypix)
			{
				left = lefts[i];
				right = rights[i];
				i++;
			} else
			{
				/* can't happen, from "if (changes)" */
				left = right = 0;

			}
			while (1)
			{
				for (; i < ypix; i++)
					if (!cdiff[i]) break;
				if (i == ypix)
					break;
				/* if there is a change in the next 3 lines,
				 * then lets block it together, but don't
				 * block more than 20 lines 
				 *
				 * note that the more lines are blocked together
				 * the wider (horiz) the block is going
				 * to potentially become
				 * */
#define XPutImage_lock(a,b,c,d,e,f,g,h,i,j) { pthread_mutex_lock(&xinput->inputlock); XPutImage(a,b,c,d,e,f,g,h,i,j); pthread_mutex_unlock(&xinput->inputlock); }
				if ( (i > last + 3) || (last - start > 20) )
				{
					XPutImage_lock(
						x11info.display,
						x11info.window,
						x11info.context,
						x11info.image,
						left,
						start,
						left,
						start,
						right - left + 1,
						last - start + 1
					);
					start = i;
					left = lefts[i];
					right = rights[i];
				} else
				{
					if (lefts[i] < left)
						left = lefts[i];
					if (rights[i] > right)
						right = rights[i];
				}
				last = i;
				i++;
			}
			
			/* send the last block */
			if (start!=ypix)
				XPutImage_lock(
					x11info.display,
					x11info.window,
					x11info.context,
					x11info.image,
					left,
					start,
					left,
					start,
					right - left + 1,
					last - start + 1
				);
		}

#if 0
		mouUpdate(&x11info); /* update cursor, also elsewhere */
#endif

		/* ms count -- for upper bound of frames-per-second */
		tick2=GetTickCount();
		if ((tick2>tick1) && (tick2-tick1<tickwait)) {
			Sleep(tickwait-tick2+tick1);
		}

	}

error:

}
